"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

# -*- coding: utf-8 -*-
from libs import base
from constants import *
from models import namespace
from utilities import connection_manager
from helpers.info_dict import InfoDict
from collections import deque
from datetime import datetime
import calendar
import time
import Queue


class Node(object):
    """Node contains all information about node"""

    def __new__(cls, address, cluster_id, validity=True):

        ip, port = address.split(":")
        #if base.is_node_available(ip, port) is False:
        #    return None
        return super(Node, cls).__new__(cls, address, cluster_id, validity)

    def __init__(self, address, cluster_id, validity=True):

        ip, port = address.split(":")

        self._ip = ip
        self._port = port
        self._address = "%s:%s" % (ip, port)
        self._node_id = base.get_node_id(ip, port)
        self._cluster_id = cluster_id

        self._status = "on"
        if base.is_node_available(ip, port) is False:
            self._status = "off"

        self._socket_queue = Queue.Queue(maxsize=SOCKET_QUEUE_SIZE)
        self._socket_empty_flag = False
        if self._status is "on":
            self._socket_queue = connection_manager.create_conn_queue(ip, port, int(SOCKET_QUEUE_SIZE))
            self._socket_empty_flag = True
        self._cluster_visibility = True

        self._id = None
        self._latency = None
        self._bins_info = None
        self._statistics = {}
        self._modified_statistics = {}
        self._configuration = {}
        self._modified_configuration = {}
        self._validity = validity

        self._namespace_list = set()
        self._namespaces = {}
        sock = self.get_socket_from_queue()
        self._build = base.get_node_build(self._ip, self._port)
        if self._build is None:
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                self._build = base.get_node_build(self._ip, self._port, sock=sock)
        self.put_socket_in_queue(sock)
        timestamp = datetime.utcnow()
        timestamp = int(calendar.timegm(timestamp.timetuple(
            )) * 1e3 + timestamp.microsecond / 1e3)
        self._read_tps_history = deque([dict(x=timestamp, secondary=None, y=None)], maxlen=TPS_HISTORY_LIMIT)
        self._write_tps_history = deque([dict(x=timestamp, secondary=None, y=None)], maxlen=TPS_HISTORY_LIMIT)

    def get_address(self):

        return self._address

    def get_socket_from_queue(self):

        #import os
        #result = os.popen("netstat -n | grep 'tcp\|udp' | awk '{print $5;}' | cut -d: -f1| sort | uniq -c | sort -nr")
        #data = result.readlines()
        #string = [line for line in data if str(self._ip) in line]
        #value = " ".join(string[0].split())

        if self._status is 'off' and self._socket_queue.empty():
           return None
        if self._status is 'on' and self._socket_queue.empty():
            self._socket_queue = connection_manager.create_conn_queue(self._ip, self._port, int(SOCKET_QUEUE_SIZE))
        while self._socket_queue.empty():
            pass

        sock = self._socket_queue.get()

        return sock
        #return self._socket_queue.get()

    def put_socket_in_queue(self, sock):

        if sock is not None:
            self._socket_queue.put(sock)

    def get_status(self):

        if base.is_node_available(self._ip, self._port):
            self._status = "on"
        else:
            self._status = "off"
        return self._status

    def set_status(self, status):

        self._status = status

    def get_cluster_visibility(self):

        return self._cluster_visibility

    def set_cluster_visibility(self, cluster_visibility):

        self._cluster_visibility = cluster_visibility

    def get_id(self):

        if self._id is None:
            sock = self.get_socket_from_queue()
            node_id = base.get_node_id(self._ip, self._port, sock=sock)
            if node_id is None:
                new_sock = connection_manager.create_socket(self._ip, self._port)
                if new_sock is not None:
                    sock = new_sock
                    node_id = base.get_node_id(self._ip, self._port, sock=sock)
            self._id = node_id
            self.put_socket_in_queue(sock)
        return self._id

    def get_build(self):

        return self._build

    def update_build(self):

        sock = self.get_socket_from_queue()
        self._build = base.get_node_build(self._ip, self._port, sock=sock)
        if self._build is None:
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                self._build = base.get_node_build(self._ip, self._port, sock=sock)
        self.put_socket_in_queue(sock)
        return self._build

    def get_cluster_key(self):

        sock = self.get_socket_from_queue()
        statistics = base.get_node_statistics(self._ip, self._port, sock=sock)
        if statistics.get('node_status', None) is 'off':
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                statistics = base.get_node_statistics(self._ip, self._port, sock=sock)
        self.put_socket_in_queue(sock)
        return statistics.get("cluster_key", UNABLE_TO_CONNECT)

    def get_statistic(self, attr):

        return self._statistics.get(attr, NOT_SUPPORTED)

    def get_statistics(self):

        return self._statistics

    def get_modified_statistics(self):

        return self._modified_statistics

    def get_read_tps_history(self, start_time, end_time):

        if start_time is not None and end_time is not None:
            return filter(lambda tps: tps['x'] > start_time and tps['x'] < end_time, list(self._read_tps_history))
        elif start_time is not None:
            return filter(lambda tps: tps['x'] > start_time, list(self._read_tps_history))
        elif end_time is not None:
            return filter(lambda tps: tps['x'] < end_time, list(self._read_tps_history))
        return self._read_tps_history

    def get_write_tps_history(self, start_time, end_time):

        if start_time is not None and end_time is not None:
            return filter(lambda tps: tps['x'] > start_time and tps['x'] < end_time, list(self._write_tps_history))
        elif start_time is not None:
            return filter(lambda tps: tps['x'] > start_time, list(self._write_tps_history))
        elif end_time is not None:
            return filter(lambda tps: tps['x'] < end_time, list(self._write_tps_history))
        return self._write_tps_history

    def get_read_tps(self):

        return self._read_tps_history[-1]

    def get_write_tps(self):

        return self._write_tps_history[-1]

    def set_invalid(self):

        self._validity = False

    def set_valid(self):

        self._validity = True

    def set_unknown(self):

        self._validity = None

    def is_valid(self):

        return self._validity

    def get_updated_statistics(self, cluster):

        old_stat_write_reqs = self.get_statistic("stat_write_reqs")
        old_stat_write_success = self.get_statistic("stat_write_success")
        old_stat_read_reqs = self.get_statistic("stat_read_reqs")
        old_stat_read_success = self.get_statistic("stat_read_success")

        sock = self.get_socket_from_queue()

        statistics = base.get_node_statistics(self._ip, self._port, sock=sock)
        if statistics.get('node_status', None) is 'off':
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                statistics = base.get_node_statistics(self._ip, self._port, sock=sock)

        self.put_socket_in_queue(sock)
        configuration = self.get_updated_configurations()
        self._modified_configuration = dict(set(configuration.items()) - set(self._configuration.items()))
        self._configuration = configuration
        node_cluster_key = statistics.get("cluster_key", None)
        if node_cluster_key == cluster.get_cluster_key():
            self.set_valid()
        elif self.is_valid() is None and node_cluster_key != None and node_cluster_key != cluster.get_cluster_key():
            self.set_invalid()
            return
        self._modified_statistics = dict(set(statistics.items()) - set(self._statistics.items()))
        self._statistics = statistics

        timestamp = datetime.utcnow()
        timestamp_posix = int(calendar.timegm(timestamp.timetuple())*1e3 + timestamp.microsecond/1e3)

        new_stat_write_reqs = self.get_statistic("stat_write_reqs")
        new_stat_write_success = self.get_statistic("stat_write_success")
        new_stat_read_reqs = self.get_statistic("stat_read_reqs")
        new_stat_read_success = self.get_statistic("stat_read_success")

        if True not in map(lambda x: x == NOT_SUPPORTED, [new_stat_write_reqs, old_stat_write_reqs, new_stat_write_success, old_stat_write_success]):
            total_write_tps = int(new_stat_write_reqs) - int(old_stat_write_reqs)
            success_write_tps = int(new_stat_write_success) - int(old_stat_write_success)
            old_timestamp = self._write_tps_history[-1]['x']
            time_delta = timestamp - datetime.utcfromtimestamp(old_timestamp/1000.0)
            time_difference = (time_delta.days * 86400) + (time_delta.seconds) + (time_delta.microseconds / 1000000.0)
            total_write_tps = total_write_tps / time_difference
            total_write_tps = round(total_write_tps, 1) if total_write_tps < 1 else round(total_write_tps)
            success_write_tps = success_write_tps / time_difference
            success_write_tps = round(success_write_tps, 1) if success_write_tps < 1 else round(success_write_tps)
            self._write_tps_history.append(dict(x=timestamp_posix, secondary=total_write_tps, y=success_write_tps))
        else:
            self._write_tps_history.append(dict(x=timestamp_posix, secondary=None, y=None))

        if True not in map(lambda x: x == NOT_SUPPORTED, [new_stat_read_reqs, old_stat_read_reqs, new_stat_read_success, old_stat_read_success]):
            total_read_tps = int(new_stat_read_reqs) - int(old_stat_read_reqs)
            success_read_tps = int(new_stat_read_success) - int(old_stat_read_success)
            old_timestamp = self._read_tps_history[-1]['x']
            time_delta = timestamp - datetime.utcfromtimestamp(old_timestamp/1000.0)
            time_difference = (time_delta.days * 86400) + (time_delta.seconds) + (time_delta.microseconds / 1000000.0)
            total_read_tps = total_read_tps / time_difference
            total_read_tps = round(total_read_tps, 1) if total_read_tps < 1 else round(total_read_tps)
            success_read_tps = success_read_tps / time_difference
            success_read_tps = round(success_read_tps, 1) if success_read_tps < 1 else round(success_read_tps)
            self._read_tps_history.append(dict(x=timestamp_posix, secondary=total_read_tps, y=success_read_tps))
        else:
            self._read_tps_history.append(dict(x=timestamp_posix, secondary=None, y=None))
        self._statistics['cluster_visibility'] = self.get_cluster_visibility()

        ##    * Unit test cases *
        ##from random import randrange

        ##free_pct_disk = ["2", "5", "10"]
        ##self._statistics['free-pct-disk'] = free_pct_disk[randrange(0, 3)]

        ##client_connections = ["90", "95", "96"]
        ##self._statistics['client_connections'] = client_connections[randrange(0, 3)]   def get_updated_statistics(self):


       ##client_connections = ["90", "95", "96"]
        ##self._statistics['client_connections'] = client_connections[randrange(0, 3)]
        ##self._statistics['proto-fd-max'] = "100"

        ##queue = ["10001", "9999"]
        ##self._statistics['queue'] = queue[randrange(0, 2)]

        ##cluster_visibility = [True, False]
        ##self._statistics['cluster_visibility'] = cluster_visibility[randrange(0, 2)]

        self.update_build()
        return self._statistics

    def get_subset_of_statistics(self, attrs=[]):

        subset = InfoDict()
        for attr in attrs:
            subset[attr] = self._statistics.get(attr, NOT_SUPPORTED)
        return subset

    def get_disk(self):

        return self.get_statistic("used-bytes-disk"), self.get_statistic("total-bytes-disk")

    def get_disk_stats(self):

        attrs = ["used-bytes-disk", "total-bytes-disk"]
        stats = self.get_subset_of_statistics(attrs)
        try:
            stats["used-bytes-disk"] = int(stats["used-bytes-disk"])
            stats["total-bytes-disk"] = int(stats["total-bytes-disk"])
            stats["free-bytes-disk"] = stats["total-bytes-disk"] - stats["used-bytes-disk"]
        except:
            stats["free-bytes-disk"] = NOT_SUPPORTED
        return stats

    def get_memory(self):

        return self.get_statistic("used-bytes-memory"), self.get_statistic("total-bytes-memory")

    def get_memory_stats(self):

        attrs = ["used-bytes-memory", "total-bytes-memory"]
        stats = self.get_subset_of_statistics(attrs)
        try:
            stats["used-bytes-memory"] = int(stats["used-bytes-memory"])
            stats["total-bytes-memory"] = int(stats["total-bytes-memory"])
            stats["free-bytes-memory"] = stats["total-bytes-memory"] - stats["used-bytes-memory"]
        except:
            stats["free-bytes-memory"] = NOT_SUPPORTED
        return stats

    def get_bins_info(self):

        sock = self.get_socket_from_queue()
        self._bins_info = base.get_node_bins_info(self._ip, self._port, sock=sock)
        if self._bins_info is None:
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                self._bins_info = base.get_node_bins_info(self._ip, self._port, sock=sock)
        self.put_socket_in_queue(sock)
        return self._bins_info

#    def get_latency(self):

#        self._latency = base.get_node_latency(self._ip, self._port)
#        return self._latency

    def get_throughput(self):

        throughput = InfoDict(read_tps=self.get_statistic("read_tps"),
            write_tps=self.get_statistic("write_tps"))
        return throughput

    def get_migration_stats(self):

        attrs = ["migrate_msgs_sent", "migrate_msgs_recv", "migrate_progress_send",
            "migrate_progress_recv", "migrate_tx_objs", "migrate_rx_objs"]
        return self.get_subset_of_statistics(attrs)

    def get_configuration(self, attr):

        return configurations.get(attr, NOT_SUPPORTED)

    def get_configurations(self):

        return self._configuration

    def get_modified_configurations(self):

        return self._modified_configuration

    def get_updated_configurations(self):

        sock = self.get_socket_from_queue()
        configuration = base.get_node_configuration(self._ip, self._port, sock=sock)
        if configuration.get('node_status', None) is 'off':
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                configuration = base.get_node_configuration(self._ip, self._port, sock=sock)

        self.put_socket_in_queue(sock)
        self._modified_configuration = dict(set(configuration.items()) - set(self._configuration.items()))
        self._configuration = configuration
        return self._configuration

    def get_subset_of_configurations(self, attrs=[]):

        subset = InfoDict()
        configurations = self.get_configurations()

        if configurations is None:
            return {}

        for attr in attrs:
            subset[attr] = configurations.get(attr, NOT_SUPPORTED)
        return subset

    def get_namespaces(self):

        return self._namespaces

    def get_namespace(self, name):

        return self._namespaces.get(name, None)

    def get_node_list(self):

        if not self.is_valid():
            return []
        sock = self.get_socket_from_queue()
        self._node_list = base.get_cluster_nodes(self._ip, self._port, sock=sock)
        if not self._node_list:
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                self._node_list = base.get_cluster_nodes(self._ip, self._port, sock=sock)
        self.put_socket_in_queue(sock)
        return self._node_list

    def get_namespace_list(self):

        if not self.is_valid():
            return []
        sock = self.get_socket_from_queue()
        namespace_list = base.get_namespaces(self._ip, self._port, sock=sock)
        if type(namespace_list) is int:
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                namespace_list = base.get_namespaces(self._ip, self._port, sock=sock)


        self.put_socket_in_queue(sock)

        if namespace_list == UNABLE_TO_CONNECT:
            return self._namespace_list

        new_namespaces = namespace_list - self._namespace_list
        self.add_namespaces(new_namespaces)

        deleted_namespaces = self._namespace_list - namespace_list
        self.delete_namespaces(deleted_namespaces)

        self._namespace_list = namespace_list
        return self._namespace_list

    def add_namespaces(self, namespace_names):

        for namespace_name in namespace_names:
            self.add_namespace(namespace_name)

    def add_namespace(self, namespace_name):

        self._namespaces[namespace_name] = namespace.Namespace(namespace_name, self._ip, self._port, self._cluster_id)

    def delete_namespaces(self, namespace_names):

        for namespace_name in namespace_names:
            self.delete_namespace(namespace_name)

    def delete_namespace(self, namespace_name):

        del self._namespaces[namespace_name]

    def get_scan_jobs(self):

        sock = self.get_socket_from_queue()
        scanned_jobs = base.get_scan_jobs(self._ip, self._port, sock=sock)
        if type(scanned_jobs) is int or not scanned_jobs:
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                scanned_jobs = base.get_scan_jobs(self._ip, self._port, sock=sock)
        self.put_socket_in_queue(sock)
        return scanned_jobs

    def get_jobs(self):

        try:
            sock = self.get_socket_from_queue()
            job_details = base.get_job_details(self._ip, self._port, sock=sock)
            if type(job_details) is int: #or not job_details
                new_sock = connection_manager.create_socket(self._ip, self._port)
                if new_sock is not None:
                    sock = new_sock
                    job_details = base.get_job_details(self._ip, self._port, sock=sock)
            self.put_socket_in_queue(sock)
            return job_details
        except:
            return []

    def get_udfs(self):

        sock = self.get_socket_from_queue()
        udf_list = base.get_udf_list(self._ip, self._port, sock=sock)
        if type(udf_list) is int or not udf_list:
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                udf_list = base.get_udf_list(self._ip, self._port, sock=sock)
        self.put_socket_in_queue(sock)
        return udf_list

    def close_sockets(self):

        connection_manager.close_sockets_in_queue(self._socket_queue)

    def replace_socket_queue(self):

        if self._status is "on":
            self._socket_queue = connection_manager.create_conn_queue(self._ip, self._port, int(SOCKET_QUEUE_SIZE))
            
    def get_node_id(self):
        return self._node_id

    def set_address(self, address):
        self._ip, self._port = address.split(":")
        self._address = address
